//
//	arfgen.cc: create dummy Attitude History File
//
//	Create a dummy AHF which starts at 1999-08-28T08:05:00 and ends
//	at 1999-08-28T12:00:00 during which the (Ra/Dec/Pos) changes
//	linearly from (30/-20/-99) to (180/0/0) - the latter attitude
//	corresponds to an attitude matrix of 1.
//

#include "common.h"
#include "Dal.h"
#include "STime.h"
#include "Celestial.h"
#include "CalOalUtility.h"

using namespace CalOalUtility;

STime start("1999-08-28T08:05:00"),
	  stop("1999-08-28T12:00:00");
double diff = stop - start;

double f(double min, double max, double t)
{
	return DEG_TO_RAD* (min + (max-min)/diff * t);
}


int main()
{
	const double ra0 = 30., ra1 = 180.,
				 dec0 = -20., dec1 = 0.,
				 pos0 = -99., pos1 = 0.;

	const double granularity = 10., blockLen = 55. * 60;	// in sec
	unsigned nTicks = UNT(blockLen/granularity);

	DataSet *set = dataSetServer->open("0010_PPPPPPOOLL_SCX00000ATS.FIT",
									   DataSet::Create);
	Table *tab = set->addTable("AHF", 4*nTicks+4);
	vector<unsigned long> size; size.push_back(19);
	Column *valtime = tab->addColumn("VALTIME", Dal::DString,
									 "data valid from this time", "", size),
		   *valdur = tab->addColumn("VALDUR", Dal::Real32,
									"duration of validity");
	size[0] = 11;
	Column *viewra = tab->addColumn("VIEWRA", Dal::DString,
									"RA of viewing direction", "", size),
		   *viewdec = tab->addColumn("VIEWDECL", Dal::DString,
									 "Dec of viewing direction", "", size),
		   *astpos = tab->addColumn("ASTPOS", Dal::Real32,
									"astronomical position angle" "deg");

	ColumnData *valtimeData = valtime->data(),
			   *viewraData = viewra->data(),
			   *viewdecData = viewdec->data(),
			   *cdvaldurData = valdur->data(),
			   *cdastposData = astpos->data();
	real32	   *valdurData = cdvaldurData->real32Data(),
			   *astposData = cdastposData->real32Data();

	unsigned nRow = 0;
	for (unsigned blockNr = 0; blockNr<4; ++blockNr)
		for (unsigned n=0; n<=nTicks; ++n) {
			double toff = blockNr * 60.*60. + n * granularity;
			STime t = start + toff;
			stringColumnRowSet(valtimeData, nRow, t.toString());
			valdurData[nRow] = granularity;
			stringColumnRowSet(viewraData, nRow,
						   RightAscension(f(ra0, ra1, toff)).toString(':'));
			stringColumnRowSet(viewdecData, nRow, Declination(
							f(dec0, dec1, toff)).toString(':'));
			astposData[nRow] = RAD_TO_DEG * f(pos0, pos1, toff);
			++nRow;
		}

	delete valtimeData;
	delete viewraData;
	delete viewdecData;
	delete cdvaldurData;
	delete cdastposData;

	dataSetServer->close(set);
}
