//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	AttitudeLocator.cc	--- implementation of AttitudeLocator classes
//
//	DESCRIPTION:
//		Implementation part of AttitudeLocator classes
//
//	CHANGE HISTORY:
//		UL, 1998-08-12: 1.0.0: - first version
//		UL, 1999-01-27: 1.1.0: - minor changes to avoid warnings under
//								 gcc 2.8.1
//		UL, 1999-02-03: 1.2.0: - added implementation for
//								 SpacecraftAttitude::toEulerAngles()
//							   - modifications pertaining to change from
//								 position angle -> astronomical position angle
//		UL, 1999-03-04: 1.3.0: - upgrades related to lowmem Dal
//		UL, 1999-04-28: 1.4.0: - fatal -> error
//		UL, 1999-08-05: 1.5.0: - constructFromAhf() is more robust against
//								 time mismatches of AHF/exp start+end
//		UL, 1999-08-18: 1.6.0: - added exception handling
//		UL, 1999-11-02: 1.7.0: - instantiation of AttitudeLocator: read
//								 AHF data for the entire observation
//		UL, 2001-03-21: 1.8.0: - fixed SPR 1960: allow for temporal overlap of
//								 consecutive AHF entries
//		UL, 2001-04-02: 1.9.0: - added constructFromRAF()
//		UL, 2001-05-03: 1.10.0:- fixed SPR 2010: cannot handle duplicate
//								 VALTIME entries in AHF
//								 (background: VALTIME lacks sub-sec precision;
//								 for VALDUR<1s one gets thus duplicate
//								 concecutive VALTIMEs)
//		UL, 2001-07-23: 1.11.0:- supress extrapolation of attitude data;
//								 if extrapolation is required attitude of
//								 closest point in time is taken
//		UL, 2001-07-25: 1.11.1:- fixed minor problem in Raf::nextEntry()
//								 which caused compilation problem under Tru64
//
//	AUTHOR:
//		U. Lammers, Wed Aug 12 14:02:33 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//		[3]: Three Axes Attitude Representation Using Position Angle,
//			 XMM-MOC-TN-0109-OAD, G. Gienger
//		[4]: Roll Angle & Conversion to/from Position Angle,
//			 INT-SYS-FD-TN-0004-OAD, J.B. Palmer
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/AttitudeLocator.cc,v 1.42 2002/01/07 16:36:12 ulammers Exp $
//

#include <fstream>
#include <cassert>
#include <cmath>
#include <iomanip>

#include "AttitudeLocator.h"
#include "Interpolator.h"
#include "CalOalUtility.h"
#include "OalServer.h"
#include "Dal.h"
#include "Celestial.h"
#include "Tracer.h"
#include "errstr.h"

using namespace CalOalUtility;

//
//	---( AttitudeLocator )---------------------------------------------------
//
void AttitudeLocator::constructFromAHF(const string &ahfSetName,
									   const STime &startTime,
									   const STime &stopTime)
{
	errstr<<"Opening Attitude History File `"<<ahfSetName<<"' and reading "
			"data covering the period "<<startTime<<" - "<<stopTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	DataSet *ahfSet = dataSetServer->open(ahfSetName, DataSet::Read);
	Table *attTab = ahfSet->table(0);
	ColumnData *valTime = attTab->column("VALTIME")->data(),
			   *viewRa = attTab->column("VIEWRA")->data(),
			   *viewDec = attTab->column("VIEWDECL")->data(),
			   *cdvalDur = attTab->column("VALDUR")->data(),
			   *cdastroPos = attTab->column("ASTPOS")->data();
	real32	*valDur = cdvalDur->real32Data(),
			*astroPos = cdastroPos->real32Data();

	unsigned nMax = attTab->rows(), n = 0;
	errstr<<"AHF covers period from "<<stringColumnRow(valTime, 0)
		  <<" - "<<stringColumnRow(valTime, nMax-1)
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	if (!nMax)
		errstr<<"The Attitude History File `"<<ahfSet->qualifiedName()
			  <<"' is empty - cannot continue"
			  <<Sas::error("EmptyAHF");

	//
	//	check if start and stop times make sense
	//
	if (startTime > stopTime)
		errstr<<"Start time ("<<startTime<<") is after stop time ("
			  <<stopTime<<")"
			  <<Sas::error("StartAfterEnd");

	//
	//	find point in time just _before_ start time
	//
	while (n < nMax &&
		   STime(stringColumnRow(valTime, n)) + valDur[n] < startTime)
		++n;

	if (n >= nMax)
		errstr<<"Data in Attitude History File `"
			  <<ahfSet->qualifiedName()<<" end before start of observation "
			    "at "<<startTime
			  <<Sas::error("TimeMismatch");

	double tStop = stopTime - Xmm::missionReferenceTime,
		   t0 = STime(stringColumnRow(valTime, n)) - Xmm::missionReferenceTime,
		   t1 = t0;
	while (n < nMax  && t0 < tStop) {
		double deltaT = valDur[n],
			   overlap = 0.;
		if (n+1 < nMax) {
			t1 = STime(stringColumnRow(valTime, n+1)) -
													Xmm::missionReferenceTime;
			overlap = t0+deltaT - t1;
		}

		if (overlap > 1.e-2) {
			errstr<<"Temporal overlap of "<<overlap<<"s between AHF entries "
				  <<n<<" ("<<Xmm::missionReferenceTime+t0<<") and "
				  <<n+1<<" ("<<Xmm::missionReferenceTime+t1
				  <<") - adjusting duration of entry "<<n
				  <<" accordingly"<<Sas::warning("TemporalOverlapInAHF");
			deltaT -= overlap;
		}

		if (t0 != t1 || n == nMax-1) {
			_time.push_back(t0);
			_duration.push_back(deltaT);
			_ra.push_back(RightAscension(stringColumnRow(viewRa, n)).angle());
			_dec.push_back(Declination(stringColumnRow(viewDec, n)).angle());
			_apos.push_back(DEG_TO_RAD*astroPos[n]);
										//	in the AHF we have
										//	the 'astronomical position angle'
										//	which is the negative standard
										//	position angle
		} else
			errstr<<"Duplicate VALTIMEs in AHF records "<<n<<" and "<<n+1
				  <<" - ignoring first one"
				  <<Sas::warning("DuplicateVALTIMEs");

		++n;
		t0 = t1;
	}

	//	check why we terminated
	if (n >= nMax && t0 + valDur[n-1] < stopTime)
		errstr<<"The attitude history data end before the end of the "
				"observation at "<<stopTime
			  <<" - this might give problems later."
			  <<Sas::warning("AHFtooShort");
	else
		if (!_time.size())
		//
		//	nothing has been read - first AHF time must lie after
		//	end of of exposure
		//
			errstr<<"The first time in the attitude history file ("<<time
				  <<") lies after the end of the observation ("<<stopTime<<")"
				  <<Sas::error("TimeMismatch");

	//	tidy up
	delete valTime;
	delete viewRa;
	delete viewDec;
	delete cdvalDur;
	delete cdastroPos;
	dataSetServer->close(ahfSet);

	_dataSource = DataSourceDescriptor(DataSourceDescriptor::AHF);
} // AttitudeLocator::constructFromAHF()


class RafSet {
public:
	RafSet(const string &name) : _name(name)
	{
		_f = new ifstream(name.c_str());
		if (!*_f)
			errstr<<"Cannot open RAF file `"<<name
				  <<Sas::error("RafFileOpenError");

		string t0, t1;
		*_f>>t0;
		*_f>>t1;
		_start = STime(t0.erase(t0.length()-1, 1));
		_stop = STime(t1.erase(t1.length()-1, 1));
		skipROL();	
	}	

	bool nextEntry(double &time, SpacecraftAttitude &att) const
	{
		checkEof();

		string dum;
		//	second record
		double valTime;
		*_f>>valTime;				//	start of validity
		time = mjd2000(valTime);
		double ra, dec, apos;
		*_f>>dum;	//	dummy field
		*_f>>ra;	//	right ascension [rad]
		*_f>>dec;	//	declination [rad]
		*_f>>apos;	//	position angle [rad]
		
		skipROL();

		att = SpacecraftAttitude(ra, dec, apos);

		return !_f->eof();
	}

	const STime &start() const { return _start; }
	const STime &stop() const { return _stop; }

	~RafSet() { delete _f; _f = 0; }

private:
	ifstream *_f;
	string	 _name;
	STime	 _start, _stop;

	//
	//	skip rest-of-line
	//
	inline void skipROL() const { _f->ignore(100, '\n'); }

	//
	//	check for unexpected EOF
	//
	inline void checkEof() const {
		if (_f->eof())
			errstr<<"Unexpected EOF in "<<_name<<Sas::error("UnexpectedEOF");
	}

	//
	//	compute number of seconds since mission reference time from
	//	MJD2000 date
	//
	double mjd2000(double t) const
	{
		static double mjdY2K = STime(2000, 1, 1).modifiedJulianDay().full();

		return STime(t + mjdY2K) - Xmm::missionReferenceTime;
	}

public:
	double toMjd2000(double t) const
	{
		static double mjds = Xmm::missionReferenceTime - STime(2000, 1, 1);

		return (t + mjds)/STime::secondsPerDay;
	}
}; // RafSet


void AttitudeLocator::constructFromRAF(const string &rafSetName,
									   const STime &startTime,
									   const STime &stopTime)
{
	errstr<<"Opening Raw Attitude File `"<<rafSetName<<"' and reading "
			"data covering the period "<<startTime<<" - "<<stopTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RafSet raf(rafSetName);
	errstr<<"RAF covers period from "<<raf.start()<<" - "<<raf.stop()
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	//
	//	check if start and stop times make sense
	//
	if (startTime > stopTime)
		errstr<<"Start time ("<<startTime<<") is after stop time ("
			  <<stopTime<<")"
			  <<Sas::error("StartAfterEnd");

	//
	//	find point in time just _before_ start time
	//
	SpacecraftAttitude att;
	double time,
		   tstart = startTime - Xmm::missionReferenceTime,
		   tend = stopTime - Xmm::missionReferenceTime;
	bool startFound = false;
	while (raf.nextEntry(time, att))
		if (time >= tstart) {
			startFound = true;
			break;
		}

	if (!startFound)
		errstr<<"Could not find raw attitude data in RAF `"
			  <<rafSetName<<"' which is valid at given start time "
			  <<startTime<<Sas::error("NoAttitudeData");

	_time.push_back(time);
	_duration.push_back(0.);
	_ra.push_back(att.ra());
	_dec.push_back(att.dec());
	_apos.push_back(att.apos());
	unsigned n=0;
	double timeprev = time-.5;
	bool neof;
	do {
		if (neof = raf.nextEntry(time, att)) {
			double tdiff=time-timeprev;
			SAS_DEBUG(cout<<"New entry: "<<n<<' '<<setprecision(10)
			    <<time<<" "<<tdiff<<endl;)
			if (tdiff<=0.) {
				errstr<<"Negative time increment "<<tdiff
					  <<"s between consecutive RAF records "
					  <<n-1<<'/'<<n<<" - ignoring latter entry"
					  <<Sas::warning("NegativeTimeIncrement");
				//
				//	occasional data with negative or zero time increments in
				//	RAF; problem has been seen in early RAFs and has been
				//	reported to FD; explanation + resolution pending;
				//	ignore these rows for the moment
				continue;
			} else if (tdiff >1.)
				errstr<<"RAF data gap of "<<std::setprecision(5)<<tdiff
					  <<"s between consecutive RAF records "
					  <<n-1<<'/'<<n<<" (t="<<std::setprecision(10)
					  <<raf.toMjd2000(time)<<"mdj2000="
					  <<std::setprecision(12)<<time<<"MET)" 
					  <<Sas::warning("AttitudeDataGap");

			timeprev = time;
			++n;
			_time.push_back(time);
			_duration.push_back(0.);
			_ra.push_back(att.ra());
			_dec.push_back(att.dec());
			_apos.push_back(att.apos());
		}
	} while (neof && time < tend);

	if (time < tend)
		errstr<<"Raw Attitude File ended at "
			  <<Xmm::missionReferenceTime+time<<" before given end time "
			  <<stopTime<<Sas::warning("PrematureEndOfRawAttitudeFile");

	if (!_time.size())
		//
		//	nothing has been read - first RAF time must lie after
		//	end of of exposure
		//
		errstr<<"The first time in the Raw Attitude File lies after the end "
				"of the observation ("<<stopTime<<")"
				  <<Sas::error("RawAttitudeFileTimeMismatch");

	_dataSource = DataSourceDescriptor(DataSourceDescriptor::RAF);
} // AttitudeLocator::constructFromRAF()



void AttitudeLocator::constructFromTHF(const string &ahfSetName,
									   const STime &startTime,
									   const STime &stopTime)
{
	//	we need to access various file to get all the needed information:
	//
	//		o AHF to obtain the absolute attitude at the beginning
	//		  of the exposure
	//		o ODF Summary File to figure which OM exposure
	//		  corresponds to the current exposure of the current instrument
	//		o OM Reference Frame Auxiliary File to get the coordinates
	//		  of the roll centre
	//		o OM non-periodic HK data to find frame time and COBT of
	//		  exposure start time
	//		o THF to read the fields: ROLL, DX, DY
	//
	const OdfSet *thf = this->parent()->odf()->
							selectSet(Xmm::OM, Xmm::TRACKING_HISTORY, 1);

	assert(&ahfSetName && &thf && &startTime && &stopTime);
	errstr<<"Sorry construction from OM THF is not yet implemented"
		  <<Sas::error("NotYetImplemented");

	_dataSource = DataSourceDescriptor(DataSourceDescriptor::THF);
} // AttitudeLocator::constructFromTHF()



AttitudeLocator::AttitudeLocator()
	: OalChild<OalServer>(0), TimeAxis(0), _maxDelta(0) {}

AttitudeLocator::AttitudeLocator(const OalServer *oalEngine,
								 double maxDelta)
	: OalChild<OalServer>(oalEngine), TimeAxis(granularity),
		_maxDelta(maxDelta)
{
	STime start(oalServer.odf()->observationStartTime()),
		  stop(oalServer.odf()->observationEndTime());

	if (oalEngine->attitudeFromStr()) {
		//
		//	we want to use the Star Tracker attitude data; first try to
		//	locate RAF if that fails use AHF
		//
		const OdfSet *attSet = oalEngine->odf()->selectSet(Xmm::SC,
														Xmm::RAW_ATTITUDE, 1,
														0, Odf::EXP_OAL,
														false, true, false);
		if (!attSet) {
			errstr<<"The ODF in directory `"
				  <<oalEngine->odf()->directoryName()
				  <<"' does not contain a Raw Attitude File - trying to "
					"access Attitude History File"
				  <<Sas::message(Msg::Lib, Msg::Verbose);
			attSet = oalEngine->odf()->selectSet(Xmm::SC,
												 Xmm::ATTITUDE_HISTORY, 1,
												 0, Odf::EXP_OAL,
												 false, true, false);
			if (!attSet)
				errstr<<"The ODF `"
					  <<oalEngine->odf()->directoryName()
					  <<"' does neither contain a Raw Attitude nor an "
						"Attitude History File - cannot continue"
					  <<Sas::error("MissingAttitudeData");

			constructFromAHF(attSet->name(), start, stop);
		} else
			constructFromRAF(attSet->name(), start, stop);
	} else {
		//
		//	we want to use the OM THF
		//
		const OdfSet *thf = oalEngine->odf()->selectSet(Xmm::OM,
													Xmm::TRACKING_HISTORY, 1);
		if (!thf)
			errstr<<"The ODF in directory `"
				  <<oalEngine->odf()->directoryName()
				  <<"` does not contain an OM Tracking History file"
				  <<Sas::error("MissingSet");

		constructFromTHF(thf->name(), start, stop);
	}
} // AttitudeLocator::AttitudeLocator()


AttitudeLocator::~AttitudeLocator()
{
} // AttitudeLocator::~AttitudeLocator()


//
//	---( AttitudeLocator::attitude )-------------------------------------
//
SpacecraftAttitude AttitudeLocator::attitude(double time) const
{
	const unsigned pointsRequired = 4;

	static double lastTime = HUGE_VAL;
	static TimeAxisIndex startIndex;

	errstr<<"Trying to retrieve attitude at t="<<setprecision(15)<<time
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	if (lastTime == HUGE_VAL)
		//	first time called
		startIndex = TimeAxisIndex(this, time);

	TimeAxisIndex leftIndex(startIndex), rightIndex(startIndex);
	++rightIndex;

	double dLeft = leftIndex.distance(time),
		   dRight = rightIndex.distance(time);

	if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
		errstr<<setprecision(15)<<"Initial: d/r="<<dLeft<<" "<<dRight
			  <<Sas::message(Msg::Lib, Msg::Noisy);

	while (dLeft < 0 && dRight < 0) {
		//	too far left
		dLeft = (++leftIndex).distance(time);
		dRight = (++rightIndex).distance(time);
		if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
			errstr<<"both neg.: d/r="<<dLeft<<" "<<dRight
				  <<Sas::message(Msg::Lib, Msg::Noisy);
	}

	while (dLeft > 0 && dRight > 0) {
		//	too far right
		dLeft = (--leftIndex).distance(time);
		dRight = (--rightIndex).distance(time);
		if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
			errstr<<"both pos.: d/r="<<dLeft<<" "<<dRight
				  <<Sas::message(Msg::Lib, Msg::Noisy);
	}

	startIndex = leftIndex;
	unsigned np = 0;
	lastTime = time;

	double xp[pointsRequired], yp[pointsRequired][3];
	while (np < pointsRequired && (leftIndex.valid() || rightIndex.valid()))
		if (fabs(dLeft) < fabs(dRight)) {
			int i1 = leftIndex.majorInd(),
				i2 = leftIndex.minorInd();
			xp[np] = _time[i1] + i2 * _granularity;
			if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
				errstr<<"Taking left: "<<xp[np]
					  <<Sas::message(Msg::Lib, Msg::Noisy);
			yp[np][0] = _ra[i1];
			yp[np][1] = _dec[i1];
			yp[np][2] = _apos[i1];
			++np;
			dLeft = (--leftIndex).distance(time);
			if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
				errstr<<" new left: "<<dLeft
					  <<Sas::message(Msg::Lib, Msg::Noisy);
		} else {
			int i1 = rightIndex.majorInd(),
				i2 = rightIndex.minorInd();

			xp[np] = _time[i1] + i2 * _granularity;
			if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
				errstr<<"Taking right: "<<xp[np]
					  <<Sas::message(Msg::Lib, Msg::Noisy);
			yp[np][0] = _ra[i1];
			yp[np][1] = _dec[i1];
			yp[np][2] = _apos[i1];
			++np;
			dRight = (++rightIndex).distance(time);
			if (errHandler.verbosity(Msg::Lib) >= Msg::Noisy)
				errstr<<" new right: "<<dRight
					  <<Sas::message(Msg::Lib, Msg::Noisy);
		}

	//
	//	check if we found the right number of points
	//
	if (np < pointsRequired) {
		errstr<<"Could not find "<<pointsRequired<<" points which are "
				"close to time tag "<<time
			  <<Sas::warning("TooLittlePointsCloseToTimeTag");
		throw NoAttitudeDataAvailable();
	}

	//
	//	check that all points are not more than <maxDelta> away from
	//	the time tag
	//
	bool interpol = false;
	unsigned mini = 0;
	double mind = 1.e10,
		   delta0 = time-xp[0];
	for (unsigned n=0; n<pointsRequired; ++n) {
		double delta = time-xp[n],
			   fabsdelta = std::fabs(delta);
		if (fabsdelta > _maxDelta) {
			errstr<<"The attitude was requested for time "
				  <<setprecision(12)<<time
				  <<" which is more than "<<_maxDelta
				  <<" s outside the period for which the attitude is known "
					"(closest time is "<<delta<<" s away)."
				  <<Sas::warning("NoAttitude");
			throw NoAttitudeDataAvailable();
		}
		if (delta * delta0 < 0.)
			//	if there is a sign change we do interpolation
			interpol = true;
		if (fabsdelta < mind) {
			mini = n;
			mind = fabsdelta;
		}
	}

	//	Are we doing interpolation?
	if (interpol) {
		double radecpos[3];
		Interpolator<3>::doPoly4(xp, yp, time, radecpos);

		return SpacecraftAttitude(radecpos[0], radecpos[1], radecpos[2]);
	}

	//	otherwise return closest attitude value
	return SpacecraftAttitude(yp[mini][0], yp[mini][1], yp[mini][2]);
} // AttitudeLocator::attitude()


const AttitudeLocator &AttitudeLocator::addCommonAttributes(
												Attributable *attrib) const
{
	static const char ATT_SRC[] = "ATT_SRC";

	Attribute *att = attrib->hasAttribute(ATT_SRC)
						? attrib->attribute(ATT_SRC)
						: attrib->addAttribute(ATT_SRC,
							   "Source of attitude data (AHF|RAF|THF)");
	*att = _dataSource.toString();

	return *this;
} // AttitudeLocator::addCommonAttributes()
