//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	AuxHKaccess.h	--- Auxiliary and HK access functionality
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2001-06-15: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Fri Jun 15 17:03:50 MET DST 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/oal/src/AuxHKaccess.h,v 1.4 2002/04/26 17:11:03 ulammers Exp $
//
#ifndef AUX_HK_ACCESS
#define AUX_HK_ACCESS

#include <vector>

#include "Xmm.h"
#include "STime.h"

class OalServer;
class Table;

static const double ObtFineTimeResolution = 1./65536.;
	//	16-bit fine time component of Central On-Board Clock will tick
	//	at maximum resolution: 2^-16 sec ~ 15.3 usec

static const double EpnFineTimeResolution = 20.48e-6;	//	EPIC-EST-SP-001,
														//	page 33
static const unsigned short MAX_FTFINE =
						static_cast<unsigned short>(1./EpnFineTimeResolution);

class HKParamFinder {
public:
	HKParamFinder(const OalServer *server);

	virtual void seek(const STime &t, double deltaT = 0.,
					  double maxDelta = 2.*STime::secondsPerDay) const;

	virtual ~HKParamFinder();

protected:
	const OalServer	*	_server;
	Xmm::Instrument		_instr;
	Table *				_table;
	mutable unsigned	_index;
};


class MainPeriodicHKParamFinder : public HKParamFinder {
public:
	MainPeriodicHKParamFinder(const OalServer *server);


	//
	//	PURPOSE:
	//		Get the values of the EPIC (MOS or PN) clock reset times from the
	//		corresponding HK file.
	//
	//		The EPIC clock is reset once at the beginning of every
	//		exposure and whenever the FTCORASE value wraps-around. There
	//		can also be commanded resets during the exposure.
	//		The times of these event are available as a COBT stamp
	//		(Coarse+Fine Time) in the periodic HK file belonging to this
	//		exposure. The function reads the value of reset time corresponding
	//		to 30 sec after the start of the exposure. This is necessary
	//		because sometimes the reset pulse becomes a couple of seconds
	//		after the exposure start.
	//
	//	RETURNS:
	//		EPIC clock reset times of current exposure as list of double COBT
	//		values
	//
	std::vector<double> resetTimes() const;

	//
	//	PURPOSE:
	//
	//	RETURNS:
	//
	unsigned wrapValue() const;
};


class AdditionalPeriodicHKParamFinder : public HKParamFinder {
public:
	AdditionalPeriodicHKParamFinder(const OalServer *server);

	unsigned short chopperValue(unsigned short quadrant) const;
};

class EpnAuxDataServer {
public:
	EpnAuxDataServer(const OalServer *oals,
					 unsigned short quadrant, unsigned short ccd,
					 unsigned short chopper, unsigned wrapValue,
					 std::vector<double> resetT);

	int locateFrame(int frameNr) const;
	int nextNormalFrame(int start) const;

	inline const std::vector<int> &ftcoarse() const { return _ftcoarse; }
	inline const std::vector<unsigned short> &ftfine() const
															{ return _ftfine; }
	inline const std::vector<int> &frame() const { return _frame; }
	inline const std::vector<double> &deltaT() const { return _deltaT; }
	inline double frameTime() const { return _frameTime; }

	enum FrameStatus { NORMAL = 1, MISSING = 2, UNUSEABLE32767 = 4,
					   RESET = 8, WRONG_QUAD_CCD = 16 };
	inline const std::vector<unsigned char> &status() const { return _status; }

	double getTime(unsigned i);
	double getObt(unsigned i);	//	getTime+resets
	inline void setTime(unsigned n, double t);

	//
	//	maximum relative deviation from mean frame time allowed to
	//	consider delta time between consecutive frames an integer multiple
	//	of frame integration time
	//
	static const double maxRelativeFrameTimeDeviation = .02;
	unsigned deltaIsIntegerMultipleOfFrameTime(unsigned i) const;
	int integerOffsetFromMultipleFrameTime(double dt) const;

	const string &tableName() { return _tableName; }

	~EpnAuxDataServer();

private:
	friend class QualityLogger;
	class AnomalyCounter {
	public:
		AnomalyCounter() : value(0) {}

		void incr() { ++value; }
		void incr(int ftcoarse, unsigned short ftfine) {
			incr();
			when.push_back(pair<int, unsigned short>(ftcoarse, ftfine));
		}
		void incr(int ftcoarse, unsigned short ftfine, double deltaT) {
			incr(ftcoarse, ftfine);
			delta.push_back(deltaT);
		}

		unsigned short									value;
		std::vector<std::pair<int, unsigned short> >	when;
		std::vector<double>								delta;
	};

	struct QualityLogger {
		QualityLogger() : minDelta(1000.), maxDelta(0.) {}

		AnomalyCounter	spurious32767s,
						lateResets,
						missingFrames,
						normalResets,
						prematureIncrements,
						negativeJumps,
						positiveJumps,
						wrongQuadCcds,
						unidentifiedDeltas;
		double			minDelta,
						maxDelta;
	};

	std::vector<int>			_ftcoarse, _frame;
	std::vector<unsigned short>	_ftfine;
	std::vector<unsigned char>	_accumResets;
	std::vector<double>			_deltaT;
	std::vector<unsigned char>	_status;
	double						_frameTime;
	unsigned short				_quadrant, _ccd,
								_chopper;
	unsigned					_wrapValue;
	std::vector<double>			_resetT;
	string						_tableName;
	QualityLogger				_qualityLogger;

	void correctAnomalies();
}; // EpnAuxDataServer

class ostream;
ostream &operator<<(ostream &, const EpnAuxDataServer::AnomalyCounter &);
ostream &operator<<(ostream &, const EpnAuxDataServer::QualityLogger &);

#endif
