//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Odf.cc	--- implementation of ODF related classes
//
//	DESCRIPTION:
//		Implementation part of ODF classes
//
//	CHANGE HISTORY:
//		UL, 1997-12-12: 1.0.0: - first version
//		UL, 1998-06-08: 1.1.0: - added ODF::selectFile()
//							   - replaced code to get ODF summary file
//								 via pipe by opendir/readdir/closedir
//		UL, 1998-07-24: 2.0.0: - overhaul - extracted from old OalAtoms.cc
//		UL, 1998-09-30: 2.1.0: - added reading of proposal summary fields
//								 from ODF Summary File (target name, etc.)
//		UL, 1998-11-20: 2.2.0: - in Odf::whatMode(): check for instrument
//								 modes only
//		UL, 1999-01-29: 2.1.0: - fixes to make it compile under egcs 1.1.1
//		UL, 1999-02-03: 2.2.0: - multiple words per line in summary file:
//								 only the first one was read - fixed
//		UL, 1999-02-09: 2.3.0: - environment variable name changed:
//									ODF_DIR -> SAS_ODFDIR
//		UL, 1999-03-03: 2.4.0: - bug fix: multiple exposures were not
//							     properly handled in Odf::registerOdfFile()
//		UL, 1999-03-17: 2.5.0: - reading of proposal summary record in summary
//								 file now ODF ICD compliant (title/firstname/
//								 surname all on one line)
//		UL, 1999-04-01: 2.6.0: - bug fix in selectFile()
//		UL, 1999-05-05: 2.7.0: - improved robustness of Odf::registerOdfFile
//		UL, 1999-05-12: 2.8.0: - added extra field length check in
//								 reading of Summary File for enhanced
//								 robustness
//		UL, 1999-05-19: 2.9.0: - support non-continuous exposure sequence
//								 numbers
//		UL, 1999-08-05: 3.0.0: - bug fix in Odf::registerOdfFile causing
//								 assert-failure on multi-exposure ODFs
//		UL, 1999-08-20: 3.1.0: - added diagnostic messages
//		UL, 1999-10-21: 3.2.0: - file names suffixes reconciled with
//								 ODF ICD issue 2.2
//		UL, 1999-11-09: 3.3.0: - changed name of environment variable
//									SAS_ODFDIR => SAS_ODF
//								 old form is still supported
//		UL, 1999-11-11: 3.3.1: - generate error if orbit number is not
//								 an integer
//		UL, 2000-02-03: 3.4.0: - fixed SPR 1079: Data viewed as SLEW
//								 check for slew data was not stringent enough
//		UL, 2000-04-10: 3.5.0: - first record of ODF summary file can be
//								 "OBSERVATION" or "SLEW"
//							   - various updates to exposure selection
//								 mechanism (fixes SPRs 1126 1228)
//		UL, 2000-05-04: 3.6.0: - minor bug fix
//		UL, 2000-06-16: 3.7.0: - several modifications to Summary File
//								 reading:
//									o can now have optional PATH to
//									  data files in first row
//									o lines starting with '//' are comments
//		UL, 2000-06-22: 3.8.0: - make reading of summary file more robust
//								 (treatment of empty lines, comments, etc.)
//		UL, 2000-07-03: 3.8.1: - fix to recognize OM Imaging data again
//								 (fixes SPR 1655: OAL does not recognize OM
//								  Imaging exposures)
//		UL, 2000-07-12: 3.8.2: - extented hasSet(Xmm::DataMode) such that
//								 upon failure it also looks for the sought
//								 set in an obervation wide scope; this
//								 is to make it work with new-style ODF
//								 in which the HK data are no longer broken
//								 down per exposure
//		UL, 2000-09-05: 3.9.0: - recognize new RGS DPP Non-Periodic-HK
//								 files of type "D1H"+"D2H"
//		UL, 2000-09-25: 3.10.0:- convenient access to IPPV implemented
//							   - summary file recognition: look for
//								 *SUM.ASC as well as *SUM.SAS; if both
//								 are present take the latter
//							   - bail out if summary file contains two
//								 observation-wide components of the same
//								 type (e.g. summary file)
//		UL, 2000-10-27: 3.11.0:- can now also read skeleton summary file from
//								 ODS
//		UL, 2000-11-08: 3.12.0:- change to fix SPR 1823: pick of ODF summary
//								 file (when *SUM.SAS~ was present this was
//								 picked instead of *SUM.SAS)
//		UL, 2000-12-08: 3.13.0:- trivial changes to make optimizer silent
//		UL, 2001-03-30: 3.13.1:- RAF is now recognized as valid ODF file type
//		UL, 2001-04-18: 3.14.0:- pseudo-exposures (auxilary/HK data) are
//								 registered with obs-start/end time as
//								 time boundaries (was fixed to
//								 1958-01-01T00:00:00)
//		UL, 2001-06-07: 3.15.0:- summary file reading robust against
//								 unexpected EOF conditions
//		UL, 2001-07-30: 3.16.0:- only register science exposures with IPPV
//								 map of their base exposures
//							   - science/engineering mode determination now
//								 uses corresponding methods from caloalutils
//		UL, 2001-09-11: 3.17.0:- OdfSet::scheduled() now returns only true
//								 for S-type exposures (used to give 'true'
//								 also for type 'X')
//		UL, 2001-10-22: 3.18.0:- various changes related to fix of SPR 2168
//							   - added OdfSet::isValid(std::string &)
//		UL, 2001-12-06: 3.19.0:- moved OdfSet methods to OdfDesignators.cc
//								 of caloalutils
//		UL, 2002-01-07: 3.19.1:- IPPV::value() now properly catches string-
//								 number conversion errors
//		UL, 2002-02-04: 3.20.0:- Summary file reading: read optional
//								 CONFIGURATION section
//		UL, 2002-02-26: 3.21.0:- added 'mode' to argument list of
//								 findExposure() - part of fixing problem
//								 in SSC-SPR-2777
//		UL, 2002-04-11: 3.22.0:- fixed small problem with determination of
//								 ODF location when SAS_ODF is given just name
//								 of summary file
//		UL, 2002-04-17: 3.23.0:- currentExposure() is now sensitive to
//								 scheduled flag
//		UL, 2002-05-24: 3.24.0:- when accesing any exposure check for
//								 empty exposure list
//		UL, 2002-06-20: 3.25.0:- broke up Odf::Odf() in smaller pieces to
//								 work around out-of-mem compiler problem
//								 at -O2 on Tru64
//							   - selectSet(Xmm::DISCARDED_LINES_DATA)
//								 works now
//
//	AUTHOR:
//		U. Lammers, Mon Dec 15 16:30:06 MET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/oal/src/Odf.cc,v 1.81 2002/06/20 10:05:20 ulammers Exp $
//

#include <cstdio>
#include <algorithm>
#include <fstream>
#include <sys/types.h>
#include <dirent.h>     //  opendir, readdir, closedir

#include "Odf.h"
#include "OalServer.h"
#include "OalUtils.h"
#include "CalOalUtility.h"
#include "ExceptionErrorHandler.h"
#include "stringutil.h"
#include "errstr.h"
#include "Tracer.h"

using namespace std;
using namespace CalOalUtility;

Odf::~Odf()
{
	for (MapType::const_iterator i = _exposures.begin();
		 i != _exposures.end(); ++i)
		for (unsigned n=0; n<(*i).second.size(); ++n)
			delete ((*i).second)[n];
} // Odf::~Odf()


ostream &operator<<(ostream &os, const ProposalInfo &pi)
{
	os<<"GO name            : "<<pi.GoName<<endl
	  <<"GO institute       : "<<pi.GoInstitute<<endl
	  <<"GO address         : "<<pi.GoAddress<<endl
	  <<"GO e-mail          : "<<pi.GoEmail<<endl
	  <<"AO                 : "<<pi.AO<<endl
	  <<"science type       : "<<pi.scienceType<<endl
	  <<"target name        : "<<pi.targetName<<endl
	  <<"target RA          : "<<pi.targetRa<<endl
	  <<"target Dec         : "<<pi.targetDec<<endl
	  <<"observation len [s]: "<<pi.obsDuration<<endl
	  <<"boresight RA       : "<<pi.boresightRa<<endl
	  <<"boresight Dec      : "<<pi.boresightDec<<endl
	  <<"position angle (lo): "<<pi.posAngleConstLo<<endl
	  <<"position angle (hi): "<<pi.posAngleConstHi<<endl
	  <<"position angle (or): "<<pi.posAngleOrigRef<<endl
	  <<"EMOS 1 priority    : "<<pi.emos1Prio<<endl
	  <<"EMOS 2 priority    : "<<pi.emos2Prio<<endl
	  <<"EPN priority       : "<<pi.epnPrio<<endl
	  <<"RGS 1 priority     : "<<pi.rgs1Prio<<endl
	  <<"RGS 2 priority     : "<<pi.rgs2Prio<<endl
	  <<"OM priority        : "<<pi.omPrio<<endl;

	return os;
} // operator<<()


ostream &operator<<(ostream &os, const IPPV &ippv)
{
	if (!ippv.invalid()) {
		os<<ippv.name()<<"="<<ippv.value()<<endl;
	}
	return os;
} // operator<<()


bool IPPVCollection::hasIPPV(const string &name) const
{
	return find(name) != end();
} // IPPVCollection::hasIPPV()


IPPV IPPVCollection::getIPPV(const string &name) const
{
	IPPVCollection::const_iterator i = find(name);
	if (i == end()) {
		errstr<<"There is no IPPV `"<<name<<"' in the actual exposure"
			  <<Sas::warning("NoSuchIPPV");
		return IPPV(name, "UNDEFINED", true);
	}

	return IPPV(name, i->second);
} // IPPVCollection::getIPPV()


ostream &operator<<(ostream &os, const IPPVCollection &ippvcoll)
{
	for (IPPVCollection::const_iterator i = ippvcoll.begin();
		 i != ippvcoll.end(); ++i)
		os<<ippvcoll.getIPPV(i->first);
	return os;
} // operator<<


ExposureRecord::ExposureRecord(unsigned counter,
							   bool isReal,
							   bool scheduled,
							   ExposureType exposureType,
							   const STime &scheduledStartTime,
							   const STime &scheduledEndTime,
							   const STime &actualStartTime,
			    			   const STime &actualEndTime)
							  :
							   _counter(counter),
							   _isReal(isReal),
							   _scheduled(scheduled),
							   _exposureType(exposureType),
							   _scheduledStartTime(scheduledStartTime),
							   _scheduledEndTime(scheduledEndTime),
							   _actualStartTime(actualStartTime),
							   _actualEndTime(actualEndTime) {}


void Exposure::addSet(const OdfSet *odfSet)
{
	_sets.push_back(odfSet);
} // Exposure::addSet()


const list<const OdfSet *> &Exposure::sets() const
{
	return _sets;
} // Exposure::sets()


ostream &operator<<(ostream &os, const ExposureRecord &expRec)
{
	os<<"exposure number      : "<<expRec.counter()<<endl
	  <<"real exposure        : "<<(expRec.isReal() ? "Y" : "N")<<endl
	  <<"scheduled            : "<<(expRec.scheduled() ? "Y" : "N")<<endl
	  <<"type                 : "<<(expRec.type() == ExposureRecord::SCIENCE
									? "SCIENCE" : "CALIBRATION")<<endl
	  <<"scheduled start time : "<<expRec.scheduledStartTime()<<endl
	  <<"scheduled end time   : "<<expRec.scheduledEndTime()<<endl
	  <<"actual start time    : "<<expRec.actualStartTime()<<endl
	  <<"actual end time      : "<<expRec.actualEndTime();

	return os;
}


static string getSumFile(const string &dirName)
{
	DIR *dirp = opendir(dirName.c_str());
	if (!dirp)
		errstr<<"Attempt to access an ODF summary file in the directory `"
			  <<dirName<<"' failed because the directory does either not "
				"exist or is not readable"
			  <<Sas::error("FileNotFound");

	string sumSetNameSAS, sumSetNameASC, sumSetName;
	while (struct dirent *dp = readdir(dirp)) {
		string s(dp->d_name);
		if (s.length() >= 8)
			s.erase(0, s.length()-7);
		if (s == "SUM.SAS") {
			sumSetNameSAS = dp->d_name;
			continue;
		}
		if (s == "SUM.ASC") {
			sumSetNameASC = dp->d_name;
			continue;
		}
	}
	closedir(dirp);

	if (!sumSetNameSAS.empty()) {
		sumSetName = sumSetNameSAS;
		if (!sumSetNameASC.empty())
			errstr<<"ODF in `"<<dirName<<"' contains both an old-style "
					"ASC and a SAS-style summary file - taking the latter"
				  <<Sas::message(Msg::Lib, Msg::Verbose);
	} else
		sumSetName = sumSetNameASC;

	if (sumSetName.empty())
		errstr<<"The directory `"<<dirName<<"' does not appear to contain "
				"an ODF - did not find the summary file which ends in "
				"\"SUM.ASC\" or \"SUM.SAS\""
			  <<Sas::error("NoSummaryFile");

	return sumSetName;
} // getSumFile()


class Odf::OdfSummaryFile : public ifstream {
public:
	OdfSummaryFile(const string &name) : ifstream(name.c_str()), _name(name) {}

	void skipRestOfLine()
	{
		readline();
	} // skipRestOfLine

	template<class T>
	inline void operator>>(T &to)
	{
		string line(readline());
		if (!*this)
			errstr<<"Unexpected EOF in line #"<<_lineNr
				  <<" of ODF summary file `"<<_name<<"'"
				  <<Sas::error("UnexpectedEOF");
		strstream str; str<<line;
		str>>to;
	} // operator>>

	inline string readline()
	{
		char s[256];
		getline(s, 256, '\n'); ++_lineNr;
		string line(trim(s, true));
		while ((line.length() == 0 ||
			   (line.length()>=2 && line[0] == '/' && line[1] == '/'))
			   && *this) {
			getline(s, 256, '\n'); ++_lineNr;
			line = trim(s, true);
		}
		return line;
	} // readline

	~OdfSummaryFile() {}

private:
	static unsigned _lineNr;
	string	_name;
};

unsigned Odf::OdfSummaryFile::_lineNr = 0;

static const char *ODFENVPTR = "SAS_ODF";	//	environment variable needs to
											//	point to directory containing
											//	ODF summary file or contains
											//	the name of the summary
											//	file directly

#define INVALID(exp, act)	\
	errstr<<"The file `"<<_summaryFileName<<"' does not appear to be a valid "\
			"ODF summary file - the string `"<<act<<"' was read where `" \
		  <<exp<<"' was expected"<<Sas::error("InvalidSummaryFile")

Odf::Odf(const string &directoryName, OalServer *oalEngine)
	: OalChild<OalServer>(oalEngine), _directoryName(directoryName)
{
	//	open ODF summary file
	OdfSummaryFile sumfile(summaryFileName());
	string buff;

	//	check if we have PATH
	buff = sumfile.readline();
	vector<string> v(breakup(buff));
	if (v.size()>1 && uppercase(v[0]) == "PATH") {
		sumfile>>buff;
		if (v[1][0] == '/')
			//	absolute path to data files
			_directoryName = v[1];
		else
			//	relative path to data files
			_directoryName += '/' + v[1];
	}

	//	read OBSERVATION section
	readObservationSection(sumfile, buff);

	//	read FILES section
	readFilesSection(sumfile, buff);

	sumfile>>buff;
	bool readExposures = false;

	readGlobalConfigSection(sumfile, buff);

	if (buff == "INSTRUMENT") {
		readExposures = true;
		//	read 6 INSTRUMENT SECTIONS
		for (unsigned i=1; i<=6; ++i) {
			errstr<<"Reading instrument section "<<i<<" from ODF summary file "
					"..."
				  <<Sas::message(Msg::Lib, Msg::Verbose);
			sumfile>>buff;
			Xmm::Instrument instr = Xmm::toInstrument(buff);

			sumfile.skipRestOfLine();
			unsigned nExps;
			sumfile>>nExps;

			for (unsigned expNr = 1; expNr <= nExps; ++expNr) {
				errstr<<"Reading exposure "<<expNr<<" ..."
					  <<Sas::message(Msg::Lib, Msg::Verbose);
				sumfile>>buff;
				if (buff != "EXPOSURE")
					INVALID("EXPOSURE", buff);

				unsigned expCount;
				sumfile>>expCount;
				char schedFlag;
				sumfile>>schedFlag;
				sumfile>>buff;
				ExposureRecord::ExposureType expType =
						buff == "SCIENCE" ?	ExposureRecord::SCIENCE
										  : ExposureRecord::CALIBRATION;
				sumfile.skipRestOfLine(); sumfile.skipRestOfLine();
				sumfile>>buff; STime sStart(buff);
				sumfile>>buff; STime sEnd(buff);
				sumfile>>buff; STime aStart(buff);
				sumfile>>buff; STime aEnd(buff);

				//	read instrument configuration
				sumfile.skipRestOfLine();
				IPPVCollection ippvs;
				unsigned nParams;
				sumfile>>nParams;
				string modeStr;
				for (unsigned n=0; n<nParams; ++n) {
					string line(sumfile.readline());
					vector<string> v(breakup(line, "/= \t"));
					if (v.size() >= 2) {
						ippvs[v[0]] = v[1];
						if (modeStr.empty() && v[0].substr(0, 9) =="DATA_MODE")
							modeStr = v[1];
					}
				}

				bool scheduled = schedFlag != 'N';
				Xmm::DataMode mode = whatMode(instr, expCount, scheduled);
				if (mode != Xmm::DATA_MODE_NONE) {
					InstrumentModePair	p(instr, mode);
					unsigned expIndex = _exposures[p].size();
					Exposure *exp = new Exposure(expCount, true, scheduled,
												 expType,
												 sStart, sEnd, aStart, aEnd,
												 ippvs, expIndex);

					_exposures[p].push_back(exp);

					errstr<<"instr/mode/size: "<<Xmm::toString(instr)<<"/"
						  <<Xmm::toString(mode)<<"/"<<_exposures[p].size()
						  <<Sas::message(Msg::Lib, Msg::Verbose);
				}
			}
			sumfile>>buff;
			if (i < 5 && buff != "INSTRUMENT")
				INVALID("INSTRUMENT", buff);
		}
	}

	if (readExposures)
		for (list<OdfSet>::iterator i = _setNames.begin();
			i != _setNames.end(); ++i)
			registerOdfSet(*i);

	readProposalDataSection(sumfile, buff);
} // Odf::Odf()


string Odf::summaryFileName()
{
	string odfName = _directoryName;
	if (_directoryName.length() == 0) {
		//	no directory name has been specified - inquire environment
		char *odfPtr;
		if ((odfPtr = getenv(ODFENVPTR)))
			odfName = string(odfPtr);
		else
			errstr<<"No name for ODF summary file/ODF directory specified and "
					"environment variable `"<<ODFENVPTR
				  <<"' is not set either - cannot continue."
				  <<Sas::error("NoODFspecified");
	}

	DIR *dir = opendir(odfName.c_str());
	if (dir) {
		closedir(dir);
		_directoryName = odfName;
		_summaryFileName = _directoryName + '/' + getSumFile(_directoryName);
	} else {
		ifstream f(odfName.c_str());
		if (!f)
			errstr<<"`"<<odfName<<"' is neither the name of a directory nor "
					"the name of an ODF summary file"
				  <<Sas::error("NoOdf");
		else {
			string::size_type pos = odfName.rfind('/');
			_directoryName = pos != string::npos ? odfName.substr(0, pos)
												 : ".";
			_summaryFileName = odfName;
		}
	}

	return _summaryFileName;
} // Odf::sumFileName()


void Odf::readObservationSection(OdfSummaryFile &sumfile, std::string &buff)
{
	//	read OBSERVATION section
	buff = trim(buff);
	if (buff != "OBSERVATION" && buff != "SLEW")
		INVALID("OBSERVATION", buff);

	sumfile>>buff;
	if (buff.length() != 10)
		errstr<<"Length of field Observation-ID (`"<<buff<<"\') in ODF "
				"summary file must have length 10"
			  <<Sas::error("IllegalFieldLength");

	_proposalId = buff.substr(0, 6);
	_observationId = buff.substr(6, 4);

	_orbitNr = 0;
	sumfile>>_orbitNr;
	if (!sumfile.good() || _orbitNr==0 || _orbitNr>9999)
		//	something has gone wrong - orbit number must be a positive
		//	integer in the range [1, 9999]
		errstr<<"Could not read orbit number from ODF summary file `"
			  <<_summaryFileName<<"' - must in the range [1, 9999] but got "
			  <<_orbitNr<<Sas::error("WrongOrbitNumber");

	_isSlew = (_proposalId[0] == '9' &&
			   toNumber<unsigned>(_proposalId.substr(1, 4)) == _orbitNr &&
			   _proposalId[_proposalId.length()-1] == '0' &&
			   _observationId.length() > 0 && _observationId[0] == '0');
		//	see ODF ICD, issue 2.4, p. 11, first bullet

	sumfile>>buff; _observationStartTime = (STime)buff;
	sumfile>>buff; _observationEndTime = (STime)buff;
} // Odf::readObservationSection()


void Odf::readFilesSection(OdfSummaryFile &sumfile, std::string &buff)
{
	//	read FILES section
	sumfile>>buff;
	if (buff != "FILES")
		INVALID("FILES", buff);

	sumfile>>_sets;
	for (unsigned n=0; n<_sets; ++n) {
		sumfile>>buff;
		OdfSet odffile(_directoryName + '/' + buff);
		_setNames.push_back(odffile);
	}
} // Odf::readFilesSection()


void Odf::readGlobalConfigSection(OdfSummaryFile &sumfile, std::string &buff)
{
	//	check if there is a global CONFIGURATION section; if there is one
	//	read it
	if (buff == "CONFIGURATION") {
		sumfile.skipRestOfLine();	//	skip number of entries
		while (sumfile) {
			string line(sumfile.readline());
			vector<string> v(breakup(line, "/= \t"));
			buff = v[0];
			if (v.size() >= 2)
				_configParams[v[0]] = v[1];
			else
				break;
		}
	}
} // Odf::readGlobalConfigSection()


void Odf::readProposalDataSection(Odf::OdfSummaryFile &sumfile, string &buff)
{
	//
	//	read Proposal Summary Record
	//
	if (buff != "PROPOSAL")
		INVALID("PROPOSAL", buff);

	if (_isSlew)
		//	no PROPOSAL data in SDFs
		return;

	getline(sumfile, buff);
	_proposalInfo.GoName = buildup(breakup(buff), ' ');

	getline(sumfile, buff);				//	institute
	_proposalInfo.GoInstitute = OalUtils::uncomment(buff);

	getline(sumfile, buff);				//	address line 1
	_proposalInfo.GoAddress = OalUtils::uncomment(buff) + ", ";
	getline(sumfile, buff);				//	address line 2
	string addr2(OalUtils::uncomment(buff));
	if (addr2.length())
		_proposalInfo.GoAddress += addr2 + ", ";
	getline(sumfile, buff);				//	town/city
	_proposalInfo.GoAddress += OalUtils::uncomment(buff) + ", ";
	getline(sumfile, buff);				//	state
	string state(OalUtils::uncomment(buff));
	if (state.length())
		_proposalInfo.GoAddress += state + ", ";
	getline(sumfile, buff);				//	country
	string country(OalUtils::uncomment(buff));

	getline(sumfile, buff);				//	ZIP
	_proposalInfo.GoAddress += OalUtils::uncomment(buff) + ", ";
	_proposalInfo.GoAddress += country;

	sumfile>>_proposalInfo.GoEmail;
	sumfile>>_proposalInfo.AO;
	sumfile>>_proposalInfo.scienceType;
	getline(sumfile, buff);
	_proposalInfo.targetName = OalUtils::uncomment(buff);

	//	Ra/Dec of target
	sumfile>>_proposalInfo.targetRa;
	_proposalInfo.targetRa *= 360./24.;	//	Ra is in hours
	sumfile>>_proposalInfo.targetDec;
	//	scheduled length of observation [s]
	sumfile>>_proposalInfo.obsDuration;
	getline(sumfile, buff);		//	skip target alternative name

	//	RA of boresight
	sumfile>>_proposalInfo.boresightRa;
	_proposalInfo.boresightRa *= 360./24.;
	//	Dec of boresight
	sumfile>>_proposalInfo.boresightDec;

	//	position angle details
	sumfile>>_proposalInfo.posAngleConstLo;
	sumfile>>_proposalInfo.posAngleConstHi;
	sumfile>>_proposalInfo.posAngleOrigRef;

	//	instrument priorities
	_proposalInfo.emos1Prio = 0; sumfile>>_proposalInfo.emos1Prio;
	_proposalInfo.emos2Prio = 0; sumfile>>_proposalInfo.emos2Prio;
	_proposalInfo.epnPrio = 0; sumfile>>_proposalInfo.epnPrio;
	_proposalInfo.rgs1Prio = 0; sumfile>>_proposalInfo.rgs1Prio;
	_proposalInfo.rgs2Prio = 0; sumfile>>_proposalInfo.rgs2Prio;
	_proposalInfo.omPrio = 0; sumfile>>_proposalInfo.omPrio;

	errstr<<"Read proposal information from ODF summary file `"
		  <<_summaryFileName<<"'"<<endl<<_proposalInfo
		  <<Sas::message(Msg::Lib, Msg::Verbose);
} // Odf::readProposalData()


Odf::MapType::const_iterator Odf::findExposure(Xmm::Instrument instr,
											   Xmm::DataMode dataMode) const
{
	checkExposureList();
	InstrumentModePair p(instr, dataMode);

	return _exposures.find(p);
} // Odf::findExposure()


const Exposure *Odf::findExposure(Xmm::Instrument instr,
								  Xmm::DataMode mode,
								  unsigned exposureNr,
								  bool scheduled) const
{
	InstrumentModePair p(instr, mode);
	MapType::const_iterator it = _exposures.find(p);
	if (it != _exposures.end()) {
		const vector<Exposure *> &v = it->second;
		unsigned N = v.size();
		for (unsigned n=0; n<N; ++n)
			if (v[n]->counter() == exposureNr &&
				v[n]->scheduled() == scheduled)
					return v[n];
	}

	return 0;
} // Odf::findExposure()


unsigned Odf::exposures(Xmm::Instrument instr, Xmm::DataMode dataMode) const
{
	Odf::MapType::const_iterator i = findExposure(instr, dataMode);

	if (i != _exposures.end())
		return (*i).second.size();

	errstr<<"No exposures for instrument `"<<instr<<"' in mode `"
		  <<dataMode
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	return 0;
} // Odf::exposures()


unsigned Odf::exposures() const
{
	return exposures(parent()->instrument(), parent()->dataMode());

	return 0;
} // Odf::exposures()


void Odf::checkExposureList() const
{
	if (_exposures.empty())
		errstr<<"It appears you have not run `odfingest' on the ODF "
				"and let the environment variable "<<ODFENVPTR<<" point to "
				"the new summary file ending in \"SUM.SAS\" - cannot "
				"continue."<<Sas::error("WrongSummaryFile");
} // checkExposureList()


bool Odf::hasSet() const
{
	return selectSet(parent()->instrument(), parent()->dataMode(),
					 parent()->exposureIndex(), 0, //parent()->ccd(),
					 Odf::EXP_OAL, false, true, false);
} // Odf::hasSet()


const OdfSet *Odf::selectSet() const
{
	return selectSet(parent()->instrument(), parent()->dataMode(),
					 parent()->exposureIndex(), 0); //parent()->ccd());
} // Odf::selectSet()


bool Odf::hasSet(Xmm::DataMode dataMode) const
{
	return selectSet(dataMode, false);
} // Odf::hasSet()


const OdfSet *Odf::selectSet(Xmm::DataMode dataMode, bool generateError) const
{
	//	some data mode require special selection - a list of all special
	//	modes follows
	if (dataMode == Xmm::DISCARDED_LINES_DATA)
		return selectDiscardedLinesDataSet(generateError);

	const OdfSet *set = selectSet(parent()->instrument(), dataMode,
								  parent()->exposureNumber(), 0,
								  Odf::EXP_ODF, true, parent()->scheduled(),
								  false);
	if (set)
		return set;

	set = selectSet(parent()->instrument(), dataMode, 0, 0, Odf::EXP_ODF,
					false, true, generateError);
	if (set && set->exposureNr() != 0) {
		set = 0;
		if (generateError)
			errstr<<"No exposure for `"
				  <<Xmm::toString(parent()->instrument())<<'/'
				  <<Xmm::toString(dataMode)<<"'"<<Sas::error("NoSuchExposure");
	}

	return set;
} // Odf::selectSet()


//
//	selection of associated DLI sets requires special logic:
//	DLI exposures are generated after the science exposure with
//	their own exposure ID (unscheduled) and one exposure per CCD;
//	the exposure are only a few secs long and start before the
//	next science exposure; here we loop over all DLI exposures
//	until a match is found on the CCD number and the exposure time
//	is not later than 15mins afer the end of the associated science
//	exposure
//
const OdfSet *Odf::selectDiscardedLinesDataSet(bool generateError)
const
{
	const Exposure *exp = currentExposure();
	const STime &currEndTime = exp->actualEndTime();
	unsigned ccd = parent()->ccd();

	Odf::MapType::const_iterator i = findExposure(parent()->instrument(),
												  Xmm::DISCARDED_LINES_DATA);
	if (i != _exposures.end()) {
		unsigned N = (*i).second.size();
		for (unsigned n=0; n<N; ++n) {
			double deltaT = (*i).second[n]->actualStartTime() - currEndTime;
			if (deltaT<0. || deltaT>15.*60.)
				//	exposure must not lie before after later than 15mins
				//	after end of current exposure
				continue;

			const list<const OdfSet *> &nameList = (*i).second[n]->sets();
			list<const OdfSet *>::const_iterator p =
						find_if(nameList.begin(), nameList.end(),
								bind2nd(mem_fun_bin_op(&OdfSet::ccdNr,
										equal_to<unsigned>()), ccd));

			if (p != nameList.end())
				//	found a match
				return *p;
		}
	}

	if (generateError)
		errstr<<"Found no "<<Xmm::toString(Xmm::DISCARDED_LINES_DATA)
			  <<" data matching exposure "<<*exp
			  <<Sas::error("NoDiscardedLinesData");

	return 0;
} // Odf::selectDiscardedLinesDataSet()


bool Odf::hasSet(Xmm::Instrument instr, Xmm::DataMode dataMode) const
{
	return selectSet(instr, dataMode,
					 parent()->exposureNumber(), 0, //parent()->ccd(),
					 Odf::EXP_ODF, true, parent()->scheduled(), false);
} // Odf::hasSet()


const OdfSet *Odf::selectSet(Xmm::Instrument instr,
							  Xmm::DataMode dataMode) const
{
	const OdfSet *set = selectSet(instr, dataMode,
								  parent()->exposureIndex(), 0,
								  Odf::EXP_ODF, true, parent()->scheduled(),
								  false);
	if (set)
		return set;

	return selectSet(instr, dataMode, parent()->exposureIndex(), 0,
					 Odf::EXP_ODF);
} // Odf::selectSet()


const Exposure *Odf::exposure(Xmm::Instrument instr,
							  Xmm::DataMode dataMode,
							  unsigned n,
							  ExposureNrType expNrType,
							  bool lookForScheduledFlagMatch,
							  bool scheduled,
							  bool generateError) const
{
	Odf::MapType::const_iterator i = findExposure(instr, dataMode);
	if (i == _exposures.end())
		if (generateError)
			errstr<<"Tried to access an exposure of instrument `"
				  <<Xmm::toString(instr)
				  <<"' in mode `"<<Xmm::toString(dataMode)<<"' but there is "
					"none in ODF `"<<directoryName()<<"'"
				  <<Sas::error("NoSuchExposure");
		else
			return 0;

	if (expNrType == Odf::EXP_OAL) {
		if (n == 0 || n > (*i).second.size())
			if (generateError)
				errstr<<"Request to access exposure #"<<n<<" for instrument `"
					  <<Xmm::toString(instr)
					  <<"' in mode `"<<Xmm::toString(dataMode)<<"' but the "
						"allowed exposure number range for this "
						"instrument/mode combination in ODF `"
					  <<directoryName()<<"' is: [1, "<<(*i).second.size()<<"]"
					  <<Sas::error("NoSuchExposure");
			else
				return 0;
	} else {
		//
		//	find exposure index
		//
		if (!n)
			n = 1;	//	0 means "not applicable" - has been registered as
					//	exposure 1
		int expIndex = -1;
		for (unsigned k=0; k<(*i).second.size(); ++k)
			if ((*i).second[k]->counter() == n) {
				if (lookForScheduledFlagMatch &&
					(*i).second[k]->isReal() &&	//	need only check against
												//	schedule flag is that is
												//	applicable
					(*i).second[k]->scheduled() != scheduled)
					continue;
				expIndex = k;
				break;
			}
		if (expIndex >= 0)
			n = expIndex + 1;
		else
			if (generateError)
				errstr<<"The requested ODF exposure #"<<n<<" for instrument `"
					  <<Xmm::toString(instr)
					  <<"' in mode `"<<Xmm::toString(dataMode)<<"' is invalid."
					  <<Sas::error("NoSuchExposure");
			else
				return 0;
	}

	return ((*i).second)[n-1];
} // Odf::exposure()


unsigned Odf::sets(Xmm::Instrument instr, Xmm::DataMode dataMode,
				   unsigned exposureNr) const
{
	InstrumentModePair p(instr, dataMode);
	Odf::MapType::const_iterator i = _exposures.find(p);

	if (i != _exposures.end() && (*i).second.size() >= exposureNr)
		return ((*i).second[exposureNr-1])->sets().size();

	return 0;
} // Odf::sets()


const Exposure *Odf::currentExposure() const
{
	return exposure(parent()->instrument(), parent()->dataMode(),
					parent()->exposureNumber(), EXP_ODF,
					true, parent()->scheduled(), true);
} // Odf::currentExposure()


const OdfSet *Odf::selectSet(Xmm::Instrument instr,
								 Xmm::DataMode dataMode,
								 unsigned exposureNr, unsigned ccdNr,
								 ExposureNrType expNrType,
								 bool lookForScheduledFlagMatch,
								 bool scheduled,
								 bool generateError) const
{
	const Exposure *exp = exposure(instr, dataMode, exposureNr, expNrType,
								   lookForScheduledFlagMatch, scheduled,
								   generateError);
	if (!exp)
		if(!generateError) {
			return 0;
		} else
			errstr<<"The ODF in directory `"<<directoryName()
				  <<"' does not an exposure: "
				  <<"instrument="<<Xmm::toString(instr)
				  <<"/dataMode="<<Xmm::toString(dataMode)
				  <<"/exposureNr="<<exposureNr<<Sas::error("NoSuchExposure");

	const list<const OdfSet *> &nameList = exp->sets();
	if (ccdNr) {
		//	look for match in CCD/node nr
		for (list<const OdfSet *>::const_iterator it =nameList.begin();
			 it != nameList.end(); ++it)
			if ((*it)->ccdNr() == ccdNr) {
				 return *it;
			}
		if (generateError)
			errstr<<"The ODF in directory `"<<directoryName()
				  <<"' does not contain a file corresponding to the request: "
				  <<"instrument="<<Xmm::toString(instr)
				  <<"/dataMode="<<Xmm::toString(dataMode)
				  <<"/exposureNr="<<exposureNr
				  <<"/ccdNr="<<ccdNr
				  <<Sas::error("NoSuchSet");
		else
			return 0;
	}

	//	ccd # not applicable - there should be only one file
	//	in the list
	return *nameList.begin();
} // Odf::selectSet()


const OdfSet *Odf::selectSet(const string &setName) const
{
	string setNameCopy(setName);
	string::size_type pos = setName.rfind("/");
	if (pos != string::npos)
		setNameCopy = setName.substr(pos+1, setName.length() - pos - 1);

	vector<string> v(breakup(setNameCopy, "()"));
	if (v.size() == 2 &&
		(uppercase(v[0]) == "ODF" || uppercase(v[0]) == "OAL")) {
		//	this is the form 'odf(...)' or 'oal(...)'
		vector<string> v2(breakup(v[1], ", \t:;"));
		unsigned expNr = 1, ccdNr = 0;
		bool scheduledFlagPresent = false,
			 scheduled = true;
		if (v2.size() == 4 || v2.size() == 5) {
			//	This the form 'ODF(instr, mode, expNr, ccdNr, [schedFlag])'
			expNr = toNumber<unsigned>(v2[2]),
			ccdNr = toNumber<unsigned>(v2[3]);
			scheduledFlagPresent = v2.size() == 5,
			scheduled = scheduledFlagPresent ? (toupper(v2[4][0]) == 'S')
											 : true;
		} else if (v2.size() != 2)
			errstr<<"ODF file name specification must be in the form "
				  <<"`ODF(instr, mode [,expnr, ccdnr, [scheduledFlag]])'"
				  <<Sas::error("IllegalNameSpecifcation");
		return selectSet(Xmm::toInstrument(v2[0]),
						     Xmm::toDataMode(v2[1]), expNr, ccdNr,
							 uppercase(v[0]) == "ODF" ? EXP_ODF : EXP_OAL,
							 scheduledFlagPresent, scheduled);
	} else {
		OdfSet odfSet(setNameCopy);
		if (odfSet.isValid()) {
			return selectSet(odfSet.instrument(),
								 odfSet.dataMode(),
								 odfSet.exposureNr(),
								 odfSet.ccdNr(), Odf::EXP_ODF);
		} else {
			return 0;
		}
	}

	return 0;
} // Odf::selectSet()


string Odf::expandSetName(const string &setName) const
{
	string s(setName);
	vector<string> v(breakup(s, "()"));
	if (v.size() == 2 && (uppercase(v[0]) == "ODF" ||
						  uppercase(v[0]) == "OAL")) {
		const OdfSet *odfSet = selectSet(setName);
		if (odfSet)
			return odfSet->name();
	}

	return s;
} // expandSetName()


void Odf::registerOdfSet(OdfSet &odfSet)
{
	InstrumentModePair	p(odfSet.instrument(), odfSet.dataMode());
	Odf::MapType::iterator i = _exposures.find(p);
	unsigned expNr = odfSet.exposureNr();
	bool scheduled = odfSet.scheduled();

	//
	//	find exposure index
	//
	int expIndex = -1;
	if (i != _exposures.end()) {
		if (expNr == 0) {
			//	this is an error: exposure-0 components apply to the entire
			//	observation, so, there can only be one
			const OdfSet *set = *i->second[0]->sets().begin();
			errstr<<"There can only be one ODF component of type "
				  <<Xmm::toString(p.first)<<'/'<<Xmm::toString(p.second)
				  <<" and `"<<set->name()<<"' has already been "
					"registered - hence, cannot register `"
				  <<odfSet.name()<<"'"<<Sas::error("DuplicateODFcomponent");
		}

		for (unsigned n=0; n<i->second.size(); ++n)
			if (i->second[n]->counter() == expNr &&
				i->second[n]->scheduled() == scheduled) {
				expIndex = n;
				break;
			}
	}

	if (expIndex < 0) {
		//
		//	exposure is not registered yet - do it
		//
		ExposureRecord::ExposureType expType = ExposureRecord::SCIENCE;
		STime sStart = observationStartTime(),
			  sEnd	 = observationEndTime(),
			  aStart = observationStartTime(),
			  aEnd	 = observationEndTime();
		Xmm::DataMode mode = whatMode(odfSet.instrument(),
									  odfSet.exposureNr(),
									  odfSet.scheduled());
		if (mode != Xmm::DATA_MODE_NONE) {
			const Exposure *exp = exposure(odfSet.instrument(), mode,
										   odfSet.exposureNr(), EXP_ODF);
			if (exp) {
				expType = exp->type();
				sStart = exp->scheduledStartTime();
				sEnd = exp->scheduledEndTime();
				aStart = exp->actualStartTime();
				aEnd = exp->actualEndTime();
			}
		}

		IPPVCollection ippvmap;
		Xmm::Instrument ins = odfSet.instrument();
		unsigned expNr = odfSet.exposureNr();
		bool scheduled = odfSet.scheduled();
		//
		//	figure out base exposure
		//
		const Exposure *baseexp = findExposure(ins, mode, expNr, scheduled);
		const vector<DataMode> &sm =
								xmmPayLoad.instrument(ins).scienceDataModes();
		Exposure *exp = new Exposure(expNr ? expNr : 1,
									 odfSet.scheduleFlagApplicable(),
									 odfSet.scheduled(),
									 expType,
									 sStart, sEnd, aStart, aEnd,
									 baseexp && find(sm.begin(), sm.end(),
													 mode) != sm.end()
										//	there is a base exposure and
										//	this is a science exposure;
										//	register with IPPV map of base
										//	exposure
										? baseexp->_ippvCollection
										//	there is no base exposure or
										//	this is no science exposure;
										//	register with empty IPPV map
										: ippvmap);
		_exposures[p].push_back(exp);
		i = _exposures.find(p);
		expIndex = _exposures[p].size()-1;
	}

	assert(expIndex >= 0 && UNT(expIndex) < _exposures[p].size());

	//
	//	register the file
	//
	(*i).second[expIndex]->addSet(&odfSet);
} // Odf::registerOdfSet


Xmm::DataMode Odf::whatMode(Xmm::Instrument instr,
							unsigned exposureNr, bool scheduled)
const
{
	DataMode mode = Xmm::DATA_MODE_NONE;
	const vector<DataMode>
		&scimodevec = xmmPayLoad.instrument(instr).scienceDataModes(),
		&modevec = xmmPayLoad.instrument(instr).scienceEngineeringDataModes();

	vector<DataMode> modes;
	for (list<OdfSet>::const_iterator i = _setNames.begin();
		 i != _setNames.end(); ++i)
		if ((*i).instrument() == instr && (*i).exposureNr() == exposureNr &&
			(*i).scheduled() == scheduled) {
			Xmm::DataMode dm = (*i).dataMode();

			//	check if this is a science mode
			if (find(scimodevec.begin(), scimodevec.end(), dm)
				!= scimodevec.end())
				return dm;

			//	check if it is a science or engineering data mode
			if (find(modevec.begin(), modevec.end(), dm) != modevec.end())
					mode = dm;
		}

	return mode;
} // Odf::whatMode()


Xmm::DataMode Odf::whatMode(Xmm::Instrument instr, unsigned exposureNr) const
{
	DataMode mode = Xmm::DATA_MODE_NONE;
	const vector<DataMode>
		&scimodevec = xmmPayLoad.instrument(instr).scienceDataModes(),
		&modevec = xmmPayLoad.instrument(instr).scienceEngineeringDataModes();

	vector<DataMode> modes;
	for (list<OdfSet>::const_iterator i = _setNames.begin();
		 i != _setNames.end(); ++i)
		if ((*i).instrument() == instr && (*i).exposureNr() == exposureNr) {
			Xmm::DataMode dm = (*i).dataMode();

			//	check if this is a science mode
			if (find(scimodevec.begin(), scimodevec.end(), dm)
				!= scimodevec.end())
				return dm;

			//	check if it is a science or engineering data mode
			if (find(modevec.begin(), modevec.end(), dm) != modevec.end())
					mode = dm;
		}

	return mode;
} // Odf::whatMode()


ostream &operator<<(ostream &os, const Odf &odf)
{
	os<<"orbit number           :"<<odf._orbitNr<<endl;
	os<<"observation start time :"<<odf._observationStartTime<<endl;
	os<<"observation end time   :"<<odf._observationEndTime<<endl;
	os<<"number of files        :"<<odf._sets<<endl;
	for (list<OdfSet>::const_iterator i = odf._setNames.begin();
		 i != odf._setNames.end(); ++i)
		os<<(*i)<<endl;

	os<<"global configuration parameters:\n";
	os<<odf.configurationParams()<<endl;

	for(Odf::MapType::const_iterator i = odf._exposures.begin();
		i != odf._exposures.end(); ++i)
	 	for (unsigned n=0; n<(*i).second.size(); ++n) {
	 		os<<*(*i).second[n]<<endl;
			os<<"IPPV:\n";
			os<<(*i).second[n]->ippvCollection()<<endl;
	}

	os<<"Proposal information:\n"<<odf._proposalInfo<<endl;

	return os;
} // operator<<


bool Odf::hasBadPixelMap() const
{
	return selectSet(parent()->instrument(),
					 Xmm::BRIGHT_PIXEL_HOUSEKEEPING, 1,
					 0, Odf::EXP_OAL, false, true, false);
} // Odf::hasBadPixelMap()


BadPixelMap Odf::badPixelMap() const
{
	const OdfSet *bps = selectSet(parent()->instrument(),
								  Xmm::BRIGHT_PIXEL_HOUSEKEEPING, 1);
	Table *bpt = dataSetServer->open(bps->name(), DataSet::Read)->table(0);
	vector<unsigned> hbrid, column, line, pixelno;
	CalOalUtility::fillVector(bpt, "HBRID", hbrid);
	CalOalUtility::fillVector(bpt, "COLUMN", column);
	CalOalUtility::fillVector(bpt, "LINE", line);
	CalOalUtility::fillVector(bpt, "PIXELNO", pixelno);
	dataSetServer->close(bpt->parent());

	const XmmInstrument &ins = XmmPayLoad::instrument(parent()->instrument());
	unsigned hbr = ins.toHBR(parent()->ccd());

	BadPixelMap map;
	unsigned N = hbrid.size();
	for (unsigned n=0; n<N; ++n) {
		if (hbrid[n] == hbr) {
			unsigned npix = pixelno[n];
			for (unsigned i=n; i<n+npix; ++i) {
				PixelLocation loc(column[i]-ins.offsetX()-ins.underScanX()+1,
								  line[i]-ins.offsetY()-ins.underScanY()+1);
				BadPixel bp(loc, 1, BadPixel::HOT, true);
				map[loc] = bp;
			}
			break;
		}
	}

	return map;
} // Odf::badPixelMap()
